@@ -39,6 +39,7 @@
 #define ONE_TEST_FAILED_MSGID  "0eee66bf98514369bef9868327a43cf1"
 #define ONE_TEST_SKIPPED_MSGID "ca0b037012363f1898466829ea163e7d"
 #define ONE_TEST_SUCCESS_MSGID "142bf5d40e9742e99d3ac8c1ace83b36"
+#define ONE_TEST_TIMED_OUT_MSGID  "db8f25eab14a4da68ef3ab3ce4b2c0bb"
 
 typedef struct {
   GHashTable *pending_tests;
@@ -46,6 +47,7 @@ typedef struct {
 
   GCancellable *cancellable;
   GPtrArray *tests;
+  GPtrArray *failed_test_msgs;
 
   int parallel;
   int test_index;
@@ -78,9 +80,11 @@ typedef struct _Test
 
   char *name;
   char **argv;
+  char **envp;
 
   TestState state;
   TestType type;
+  guint timeout;
 } Test;
 
 typedef struct _TestClass
@@ -96,6 +100,7 @@ test_finalize (GObject *gobject)
 {
   Test *self = (Test*) gobject;
   g_strfreev (self->argv);
+  g_strfreev (self->envp);
   g_clear_object (&self->tmpdir);
   g_object_unref (self->prefix_root);
   g_object_unref (self->path);
@@ -130,6 +135,8 @@ load_test (GFile         *prefix_root,
   gs_free char *exec_key = NULL;
   gs_free char *type_key = NULL;
   const char *test_path;
+  char **env_key = NULL;
+  GError *internal_error = NULL;
 
   test = g_object_new (test_get_type (), NULL);
 
@@ -166,7 +173,15 @@ load_test (GFile         *prefix_root,
                    "Unknown test type '%s'", type_key);
       goto out;
     }
-  
+
+  env_key = g_key_file_get_string_list (keyfile, "Test", "TestEnvironment", NULL, &internal_error);
+  if (internal_error != NULL &&
+      !(internal_error->domain == G_KEY_FILE_ERROR &&
+        internal_error->code == G_KEY_FILE_ERROR_KEY_NOT_FOUND))
+    goto out;
+  else
+    test->envp = env_key;
+
   test->state = TEST_STATE_LOADED;
 
   ret = TRUE;
@@ -175,6 +190,7 @@ load_test (GFile         *prefix_root,
  out:
   if (!ret && test)
     g_prefix_error (error, "Test '%s': ", test->name);
+  g_clear_error (&internal_error);
   g_clear_pointer (&keyfile, g_key_file_free);
   return ret;
 }
@@ -182,7 +198,9 @@ load_test (GFile         *prefix_root,
 static TestRunnerApp *app;
 
 static gboolean opt_list;
+static int opt_firstroot;
 static int opt_parallel = 1;
+static int opt_cancel_timeout = 5*60;
 static char * opt_report_directory;
 static char **opt_dirs;
 static char *opt_status;
@@ -192,9 +210,11 @@ static GOptionEntry options[] = {
   { "dir", 'd', 0, G_OPTION_ARG_STRING_ARRAY, &opt_dirs, "Only run tests from these dirs (default: all system data dirs)", NULL },
   { "list", 'l', 0, G_OPTION_ARG_NONE, &opt_list, "List matching tests", NULL },
   { "parallel", 'p', 0, G_OPTION_ARG_INT, &opt_parallel, "Specify parallelization to PROC processors; 0 will be dynamic)", "PROC" },
+  { "first-root", 0, 0, G_OPTION_ARG_NONE, &opt_firstroot, "Only use first entry in XDG_DATA_DIRS", "PROC" },
   { "report-directory", 0, 0, G_OPTION_ARG_FILENAME, &opt_report_directory, "Create a subdirectory per failing test in DIR", "DIR" },
-  { "status", 0, 0, G_OPTION_ARG_STRING, &opt_status, "Output status information (yes/no/auto)", NULL },
+  { "status", 0, 0, G_OPTION_ARG_STRING, &opt_status, "Output status information", "yes/no/auto" },
   { "log-msgid", 0, 0, G_OPTION_ARG_STRING, &opt_log_msgid, "Log unique message with id MSGID=MESSAGE", "MSGID" },
+  { "timeout", 't', 0, G_OPTION_ARG_INT, &opt_cancel_timeout, "Cancel test after timeout seconds; defaults to 5 minutes", "TIMEOUT" },
   { NULL }
 };
 
@@ -280,12 +300,13 @@ log_test_completion (Test *test,
   else if (test->state == TEST_STATE_COMPLETE_FAILED)
     {
       msgid_value = ONE_TEST_FAILED_MSGID;
-      msg = g_strconcat ("FAILED: ", test->name, " (", reason, ")", NULL);
+      msg = g_strconcat ("FAIL: ", test->name, " (", reason, ")", NULL);
+      g_ptr_array_add (app->failed_test_msgs, g_strdup (msg));
     }
   else if (test->state == TEST_STATE_COMPLETE_SKIPPED)
     {
       msgid_value = ONE_TEST_SKIPPED_MSGID;
-      msg = g_strconcat ("SKIPPED: ", test->name, NULL);
+      msg = g_strconcat ("SKIP: ", test->name, NULL);
     }
   else
     g_assert_not_reached ();
@@ -297,7 +318,7 @@ log_test_completion (Test *test,
   keys[1] = msgid;
   keys[2] = NULL;
 
-  gs_log_structured (msg, (const char *const*)keys);
+  gs_log_structured_print (msg, (const char *const*)keys);
 }
 
 static void
@@ -339,6 +360,10 @@ on_test_exited (GObject       *obj,
     }
   else
     {
+      if (test->timeout) {
+        g_source_remove (test->timeout);
+        test->timeout = 0;
+      }
       test->state = TEST_STATE_COMPLETE_SUCCESS;
       log_test_completion (test, NULL);
     }
@@ -346,8 +371,12 @@ on_test_exited (GObject       *obj,
   /* Keep around temporaries from failed tests */
   if (!(failed && opt_report_directory))
     {
-      if (!gs_shutil_rm_rf (test->tmpdir, cancellable, error))
-        goto out;
+      gs_unref_object GFile *test_tmpdir_stamp = g_file_get_child (test->tmpdir, ".testtmp");
+      if (g_file_query_exists (test_tmpdir_stamp, NULL))
+        {
+          if (!gs_shutil_rm_rf (test->tmpdir, cancellable, error))
+            goto out;
+        }
     }
 
  out:
@@ -357,10 +386,10 @@ on_test_exited (GObject       *obj,
     g_task_return_boolean (task, TRUE);
 }
 
+#ifdef ENABLE_SYSTEMD_JOURNAL
 static void
 setup_test_child (gpointer user_data)
 {
-#ifdef ENABLE_SYSTEMD_JOURNAL
   Test *test = user_data;
   int journalfd;
   
@@ -375,7 +404,18 @@ setup_test_child (gpointer user_data)
         ret = dup2 (journalfd, 2);
       while (ret == -1 && errno == EINTR);
     }
+}
 #endif
+
+static gboolean
+cancel_test (gpointer data)
+{
+  gs_unref_object GSSubprocess *proc = data;
+  gs_subprocess_force_exit (proc);
+  gs_log_structured_print_id_v (ONE_TEST_TIMED_OUT_MSGID,
+                                "Test timed out after %u seconds",
+                                opt_cancel_timeout);
+  return FALSE;
 }
 
 static void
@@ -391,7 +431,6 @@ run_test_async (Test                *test,
   GError **error = &local_error;
   gs_free char *testname = NULL;
   gs_free char *test_tmpdir = NULL;
-  gs_unref_object GFile *test_tmpdir_f = NULL;
   gs_free char *test_squashed_name = NULL;
   gs_free char *test_tmpname = NULL;
   gs_unref_object GSSubprocessContext *proc_context = NULL;
@@ -425,18 +464,45 @@ run_test_async (Test                *test,
     {
       test_tmpdir = g_build_filename (opt_report_directory, test_squashed_name, NULL);
       test->tmpdir = g_file_new_for_path (test_tmpdir);
-      if (!gs_shutil_rm_rf (test_tmpdir_f, cancellable, error))
+      if (!gs_shutil_rm_rf (test->tmpdir, cancellable, error))
         goto out;
-      if (!gs_file_ensure_directory (test_tmpdir_f, TRUE, cancellable, error))
+      if (!gs_file_ensure_directory (test->tmpdir, TRUE, cancellable, error))
         goto out;
     }
 
+  /* We create a .testtmp stamp file so that tests can *know* for sure
+   * they're in a temporary directory.  This is used by at least the
+   * OSTree tests as protection against someone running a test script
+   * outside of the framework, as it might overwrite files in their
+   * source directory, etc.
+   *
+   * Also, when we do the rm -rf, we test for the file to be doubly
+   * sure that we're deleting the right tmpdir.
+   */ 
+  {
+    gs_unref_object GFile *test_tmpdir_stamp = g_file_get_child (test->tmpdir, ".testtmp");
+
+    if (!g_file_replace_contents (test_tmpdir_stamp, "", 0, NULL, FALSE, 0, NULL, cancellable, error))
+      goto out;
+  }
+
   proc_context = gs_subprocess_context_new (test->argv);
   gs_subprocess_context_set_cwd (proc_context, test_tmpdir);
+  gs_subprocess_context_set_environment (proc_context, test->envp);
+  if (!opt_report_directory)
+    {
 #ifdef ENABLE_SYSTEMD_JOURNAL
-  if (gs_stdout_is_journal ())
-    gs_subprocess_context_set_child_setup (proc_context, setup_test_child, test);
+      if (gs_stdout_is_journal ())
+        gs_subprocess_context_set_child_setup (proc_context, setup_test_child, test);
 #endif
+    }
+  else
+    {
+      const char *test_output_filename = "output.txt";
+      gs_free char *test_output_path = g_build_filename (test_tmpdir, test_output_filename, NULL);
+      gs_subprocess_context_set_stdout_file_path (proc_context, test_output_path);
+      gs_subprocess_context_set_stderr_disposition (proc_context, GS_SUBPROCESS_STREAM_DISPOSITION_STDERR_MERGE);
+    }
 
   proc = gs_subprocess_new (proc_context, cancellable, error);
   if (!proc)
@@ -445,6 +511,7 @@ run_test_async (Test                *test,
   test->state = TEST_STATE_EXECUTING;
 
   gs_subprocess_wait (proc, cancellable, on_test_exited, task);
+  test->timeout = g_timeout_add_seconds (opt_cancel_timeout, cancel_test, g_object_ref (proc));
 
  out:
   if (local_error)
@@ -619,6 +686,9 @@ main (int argc, char **argv)
   memset (&appstruct, 0, sizeof (appstruct));
   app = &appstruct;
 
+  /* avoid gvfs (http://bugzilla.gnome.org/show_bug.cgi?id=526454) */
+  g_setenv ("GIO_USE_VFS", "local", TRUE);
+
   context = g_option_context_new ("[PREFIX...] - Run installed tests");
   g_option_context_add_main_entries (context, options, NULL);
 
@@ -645,6 +715,7 @@ main (int argc, char **argv)
 
   app->pending_tests = g_hash_table_new (NULL, NULL);
   app->tests = g_ptr_array_new_with_free_func ((GDestroyNotify)g_object_unref);
+  app->failed_test_msgs = g_ptr_array_new_with_free_func ((GDestroyNotify)g_free);
 
   if (opt_dirs)
     datadirs_iter = (const char *const*) opt_dirs;
@@ -663,6 +734,9 @@ main (int argc, char **argv)
       if (!gather_all_tests_recurse (prefix_root, prefix_root, "", app->tests,
                                      cancellable, error))
         goto out;
+
+      if (opt_firstroot)
+        break;
     }
 
   if (argc > 1)
@@ -781,6 +855,11 @@ main (int argc, char **argv)
                                     ret ? "" : " (incomplete)",
                                     total_tests, n_passed, n_skipped, n_failed,
                                     rusage_str != NULL ? rusage_str : "");
+      if (gs_console_get ())
+        {
+          for (i = 0; i < app->failed_test_msgs->len; i++)
+            g_print ("%s\n", (char *) app->failed_test_msgs->pdata[i]);
+        }
     }
   g_clear_pointer (&app->pending_tests, g_hash_table_unref);
   g_clear_pointer (&app->tests, g_ptr_array_unref);
